@@ -38,7 +38,6 @@
 #include "postmaster/postmaster.h"
 #include "tcop/tcopprot.h"
 #include "utils/help_config.h"
-#include "utils/memutils.h"
 #include "utils/pg_locale.h"
 #include "utils/ps_status.h"
 #ifdef WIN32
@@ -79,15 +78,6 @@ main(int argc, char *argv[])
 	 */
 	argv = save_ps_display_args(argc, argv);
 
-	/*
-	 * Fire up essential subsystems: error and memory management
-	 *
-	 * Code after this point is allowed to use elog/ereport, though
-	 * localization of messages may not work right away, and messages won't go
-	 * anywhere but stderr until GUC settings get loaded.
-	 */
-	MemoryContextInit();
-
 	/*
 	 * Set up locale information from environment.	Note that LC_CTYPE and
 	 * LC_COLLATE will be overridden later from pg_control if we are in an
